// *************************************************************************
//    BitVector.C
// *************************************************************************

#include <stdlib.h>
#include <string.h>

#include "BitVector.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
//    Number of BitWords to represent 'i' bits

static int bit_words_needed(int i)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (i + BIT_VECTOR_WORD - 1) >> BIT_VECTOR_WORD_LOG;
}

// *************************************************************************
void Cor_BitVector::init(const Cor_BitVector& b2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int words_needed = bit_words_needed(b2.length);
    if (words_needed <= 1) {
        u.word0 = b2.u.word0;
    } else {
        u.words = new BitWord[words_needed];
        BitWord *src = b2.u.words;
        BitWord *dst = u.words;
        while (--words_needed >= 0)
            *dst++ = *src++;
    }
    length = b2.length;
}

// *************************************************************************
Cor_BitVector::~Cor_BitVector()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length > BIT_VECTOR_WORD)
        delete [] u.words;
}

// *************************************************************************
void Cor_BitVector::printon(ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char ch;
    for (int i = 0; i < length; ++i) {
      ch = '0' + test(i);
      *file << ch;
    }
}

// *************************************************************************
int Cor_BitVector::next(int i) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!this)
      return(-1);

    for (++i; i < length; ++i)
      if (test(i))
        return(i);

    return(-1);
}

// *************************************************************************
void Cor_BitVector::setlen(int l)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (l > BIT_VECTOR_WORD) {
        BitWord *old_words = nc_words();
        int old_length = bit_words_needed(length);
        int new_length = bit_words_needed(l);
        int j;
        int new_bytes = new_length * sizeof(BitWord);
        BitWord *new_words = new BitWord[new_bytes];
        memset(new_words, 0, new_bytes);

        for (j = old_length; --j >= 0; )
            new_words[j] = old_words[j];

        u.words = new_words;
        if (old_length > 1)
            delete [] old_words;

    } else if (length > BIT_VECTOR_WORD) {
        BitWord w = u.words[0];
        delete [] u.words;
        u.word0 = w;
    }
    length = l;
}

// *************************************************************************
void Cor_BitVector::set(int index, int value /*= 1*/)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (index >= length)
      setlen(index+1);
    BitWord *ptr = &nc_words()[index>>BIT_VECTOR_WORD_LOG];
    BitWord mask = 1 << (index & BIT_VECTOR_WORD-1);

    if (value)
      *ptr |= mask;
    else
      *ptr &= ~mask;
}

// *************************************************************************
void Cor_BitVector::operator=(const Cor_BitVector& b2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Ok for (&b2 == this).
    int words_needed = bit_words_needed(b2.length);
    if (words_needed <= 1) {
        if (length > BIT_VECTOR_WORD)
          delete [] u.words;
        u.word0 = b2.u.word0;
    } else {
    if (length > BIT_VECTOR_WORD){
      delete [] u.words;
      u.words = new BitWord[words_needed];
    } else
      u.words = new BitWord[words_needed];
    BitWord *src = b2.u.words;
    BitWord *dst = u.words;
    while (--words_needed >= 0)
        *dst++ = *src++;
    }
    length = b2.length;
}

// *************************************************************************
int Cor_BitVector::operator==(const Cor_BitVector& b2) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length != b2.length)
        return(0);

    const BitWord *ptr1 = words();
    const BitWord *ptr2 = b2.words();
    int words1 = bit_words_needed(length);
    int words2 = bit_words_needed(b2.length);
    int i = words1 > words2 ? words2 : words1;
    while (--i >= 0) {
        if (*ptr1++ != *ptr2++)
          return 0;
    }
    i = words1 - words2;
    if (i > 0) {
        while (--i >= 0)
          if (*ptr1++ != 0)
            return 0;
    } else if (i < 0) {
        for (i = -1; --i >= 0; ) {
          if (*ptr2++ != 0)
            return 0;
        }
    }
    return 1;
}

// *************************************************************************
int Cor_BitVector::implies(const Cor_BitVector& b2) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    const BitWord *ptr1 = words();
    const BitWord *ptr2 = b2.words();
    int words1 = bit_words_needed(length);
    int words2 = bit_words_needed(b2.length);
    int i = words1 > words2 ? words2 : words1;

    while (--i >= 0) {
      if (*ptr1++ & ~*ptr2++)
        return(0);
    }
    i = words1 - words2;
    while (--i >= 0) {
      if (*ptr1++ != 0)
        return(0);
    }
    return(1);
}

// *************************************************************************
int Cor_BitVector::count() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i=0, c;

    for(c = next(-1); c >= 0; c = next(c) )
      i++;

    return(i);
}

// *************************************************************************
void Cor_BitVector::or_bits(const Cor_BitVector *b2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int c;
    for(c = b2->next(-1); c >= 0; c = b2->next(c) )
      set(c);
}

// *************************************************************************
void Cor_BitVector::clear()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    memset(nc_words(), 0, bit_words_needed(length) * sizeof(BitWord));
}

// *************************************************************************
int Cor_BitVector::any() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    const BitWord *ptr = words();
    int i = bit_words_needed(length);
    while (--i >= 0) {
      if (*ptr++)
        return 1;
    }
    return 0;
}

// *************************************************************************
int Cor_BitVector::operator&(const Cor_BitVector& b2) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    const BitWord *ptr1 = words();
    const BitWord *ptr2 = b2.words();
    int i = length;

    if (i > b2.length)
      i = b2.length;
    i = (i+BIT_VECTOR_WORD-1) >> BIT_VECTOR_WORD_LOG;
    while (--i >= 0) {
      if (*ptr1++ & *ptr2++)
        return 1;
    }
    return 0;
}

// *************************************************************************

