// *************************************************************************
//    Context.C
// *************************************************************************

#include "Context.h"

#include "Globals.h"
#include "Unify.h"
#include "StorageRelation.h"
#include "HashIndexSet.h"
#include "Debug.h"

// *************************************************************************
Cor_Context::Cor_Context()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    all_nodes      = NULL;
    first          = NULL;
    last           = NULL;
    first_marked   = NULL;
    last_marked    = NULL;
}

// *************************************************************************
Cor_Context::~Cor_Context()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    free_all();
}

// *************************************************************************
void Cor_Context::add_new_goal(Cor_GoalNode *g_node, Cor_GoalNode *par)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // NOTE: this does not do the duplicate checking on the Context.
    // that HAS to be done elsewhere

    g_node->create_goalnode_info(par);
    Cor_ContextNode *c_node = new Cor_ContextNode(g_node, this);

    // Save all new nodes in a list so that they can be deleted later.
    if (all_nodes) {
        all_nodes->all_prev = c_node;
        c_node->all_next = all_nodes;
    } 
    all_nodes = c_node; 

    // the new goal and context are unmarked

    // first, we handle the extreme case of seed facts on Context
    // NOTE: all unmarked seed facts have a node_number of 0

    if (par == NULL) {
        // insert at end of Context
        insert_last(c_node);
        return;
    }

    // With a non-null "par" field, have to insert it just before
    // the next marked context node from context node of par
    // or at end of context if context node of par is last marked node

    ASSERT((par->is_marked()));
    Cor_ContextNode *par_c_node = par->find_context_node();
    ASSERT((par_c_node != NULL));
    ASSERT((par_c_node->is_marked()));

    // node number of newly inserted node is same as that of its parent's node
    c_node->node_number = par_c_node->node_number;

    Cor_ContextNode *par_next_c_node = par_c_node->next_marked;

    if (par_next_c_node == NULL) { // insert at end of Context
        insert_last(c_node);
    } else {
        insert_before(c_node, par_next_c_node);
    }
}

// *************************************************************************
void Cor_Context::insert_last(Cor_ContextNode *c_node)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (first == NULL) {
        // this is the first context node in Context
        first = c_node;
        last = c_node;
    } else {
        c_node->prev = last;
        last->next = c_node;
        last = c_node;
    }
}

// *************************************************************************
void Cor_Context::insert_before(Cor_ContextNode *c_node,
            Cor_ContextNode *par_next_c_node)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    c_node->next = par_next_c_node;
    c_node->prev = par_next_c_node->prev;
    par_next_c_node->prev = c_node;
    c_node->prev->next = c_node;
}

// *************************************************************************
/*
    NOTE: deleting goal nodes from a context node in the non-ground case
    can only be done by setting a delete flag.  how does that affect the
    running time?

    In the all ground case, any node which is not a singleton node
    should have no unmarked goals in it.  so for the ground case,
    we have to worry about deletion only from singleton nodes,
    which should not mess anything up.
*/

void Cor_Context::delete_goal(Cor_GoalNode *g)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // NOTE: ordered search needs to delete only unmarked goals during
    // subsumption checks

    ASSERT((!(g->is_marked())));

    Cor_ContextNode *g_c_node = g->find_context_node();
    ASSERT((g_c_node != NULL));

    // the following is a hack
    if ((g_c_node->first_marked_goal == NULL) &&
            (g_c_node->first_unmarked_goal == NULL)) {
        // something shady happened
        return;
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_ordsearch) {
            *(Cor_exEnv->trace_strm) << "\n\tTrying to delete GoalNode:\n";
            g->print_all(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "\tCorresponding ContextNode is:\n";
            g_c_node->print_all(Cor_exEnv->trace_strm);
        }
#    endif

    if (g_c_node->is_marked()) {
        // unmarked goal in marked node

        g->c_node = NULL;
        g->parent_goal = NULL;
        g->parent_in_set = NULL;

        // mark goal as deleted, do nothing else
        delete_id_tuple(g->owner_rel,g);

        // NOTE: this is safe since unmarked goal nodes in marked
        //       context nodes are guaranteed to be leaf nodes because
        //       of the set manipulation operations
    } else {
        // unmarked goal in singleton unmarked node -- can delete everything
        ASSERT((g_c_node->first_marked_goal == NULL));
        ASSERT((g_c_node->first_unmarked_goal == g_c_node->last_unmarked_goal));

        g->c_node = NULL;
        g->parent_goal = NULL;
        g->parent_in_set = NULL;

        // marking the goal itself as deleted
        delete_id_tuple(g->owner_rel,g);

        // unlink the context node
        unlink_node(g_c_node);

        g_c_node->free_all();

          // Delete from the doubly-linked list all_nodes
        if (g_c_node->all_prev)
            g_c_node->all_prev->all_next = g_c_node->all_next;
        else
            all_nodes = g_c_node->all_next;

        if (g_c_node->all_next)
            g_c_node->all_next->all_prev = g_c_node->all_prev;

        delete g_c_node;
    }
}

// *************************************************************************
void Cor_Context::unlink_node(Cor_ContextNode *g_c_node)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (g_c_node->next == NULL) {
        // last node in Context
        if (g_c_node->prev != NULL) {
            // not first node in Context
            g_c_node->prev->next = NULL;
        }
        last = g_c_node->prev;
    } else {
        // not last node in Context
        g_c_node->next->prev = g_c_node->prev;
        if (g_c_node->prev != NULL) {
            // not first node in Context
            g_c_node->prev->next = g_c_node->next;
        }
    }

    if (g_c_node->is_marked()) {
        if (g_c_node->next_marked == NULL) {
            // last marked node
            if (g_c_node->prev_marked != NULL) {
                // not first marked node
                g_c_node->prev_marked->next_marked = NULL;
            }
            last_marked = g_c_node->prev_marked;
        } else {
            // not last marked node
            g_c_node->next_marked->prev_marked = g_c_node->prev_marked;
            if (g_c_node->prev_marked != NULL) {
                // not first marked node
                g_c_node->prev_marked->next_marked = g_c_node->next_marked;
            }
        }
    }

    // update context info
    if (last == NULL)
        first = NULL;
    if (last_marked == NULL)
        first_marked = NULL;
}

// *************************************************************************
Cor_GoalNode *Cor_Context::get_unmarked_goal(int &any_new_fact_in_done)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    any_new_fact_in_done = 0;

    if (last == NULL) {
        // empty context
        if (Cor_exEnv->dbg_ordsearch) {
            *(Cor_exEnv->trace_strm) << "  OrdSearch: Empty context.\n";
        }
        return(NULL);
    }

    Cor_GoalNode *fac;
    Cor_ContextNode *c = last;
    fac = c->get_unmarked_goal();

    if (fac == NULL) {
        // node contains no unmarked subgoal
        ASSERT((c->is_marked()));
        if (Cor_exEnv->dbg_ordsearch) {
          *(Cor_exEnv->trace_strm) << "OrdSearch: moving goals to done.\n";
        }
        c->move_goals_to_done();
        any_new_fact_in_done = 1;
        unlink_node(c);

        c->free_all();

        // ERROR!!  This causes major memory overwrite problems.
        // If c is deleted here, there are still pointers to it from the
        // c_node of some GoalNode.  This gets accessed later via
        // find_context_node, and we get screwed. : PRAVEEN
        //delete c;
        // since we have moved facts to the done relation new inferences
        // can possibly be made in Ordered Search
    } else {
        // have an unmarked goal
        if (Cor_exEnv->dbg_ordsearch) {
            *(Cor_exEnv->trace_strm) << "OrdSearch: found unmarked goal: ";
            *(Cor_exEnv->trace_strm) << "    "
                    << fac->owner_rel->name.symbol() << "(";
            fac->args().print(NULL,Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ") \n";
            print_all(Cor_exEnv->trace_strm);
        }

        if (!(c->is_marked())) {
            c->mark_node();
            update_marked_info(c);

            // set node number of newly marked context node
            if (c->prev_marked == NULL) {
                c->node_number = 1;
            } else {
                c->node_number = c->prev_marked->node_number + 1;
            }
        }
    }

    return(fac);
}

// *************************************************************************
void Cor_Context::update_marked_info(Cor_ContextNode *c_node)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // we are assuming that only the highest unmarked node on
    // Context ever gets marked, which is fine for ordered search

    if (first_marked == NULL) {
        // no marked nodes in Context yet
        first_marked = c_node;
        last_marked = c_node;
    } else {
        // only marked context nodes are linked up in this way
        c_node->prev_marked = last_marked;
        c_node->next_marked = NULL;
        last_marked->next_marked = c_node;
        last_marked = c_node;
    }
}

// *************************************************************************
void Cor_Context::collapse_nodes(Cor_GoalNode *from,
        Cor_GoalNode *par, Cor_GoalNode *to)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (from == to)
        return;        // no collapsing required

    ASSERT((from->is_marked()));
    ASSERT((par->is_marked()));
    ASSERT((!(to->is_marked())));

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_ordsearch) {
            *(Cor_exEnv->trace_strm) << "OrdSearch: collapsing nodes\n";
            *(Cor_exEnv->trace_strm) << "           from:  ";
            from->args().print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "\n           par :  ";
            par->args().print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "\n           to  :  ";
            to->args().print(NULL, Cor_exEnv->trace_strm);
        }
#    endif

    Cor_ContextNode *from_c_node = from->find_context_node();
    Cor_ContextNode *par_c_node = par->find_context_node();

    if (from_c_node != par_c_node) {
        // collapse marked nodes into from_c_node
        Cor_ContextNode *c_tmp;
        Cor_ContextNode *pnxt_tmp = par_c_node->next_marked;

        for (c_tmp = from_c_node->next_marked; c_tmp && c_tmp != pnxt_tmp; ) {
            // First check if any cycle of negative dependencies exist.
            // Only the c_tmp nodes need to be checked, not the
            // from_c_node, since that could be negatively dependent on
            // a previous goal, without any problems

            // ASSERT((!(c_tmp->negated)));
            if (c_tmp->negated) {
                *(Cor_exEnv->err_strm) << 
                    "Cycle of negative dependencies detected!\n";
            }

            from_c_node->merge_nodes(c_tmp);
            unlink_node(c_tmp);

            delete c_tmp;
            c_tmp = from_c_node->next_marked;
        }
    }

    // now collapse "to" into "from"
    Cor_ContextNode *to_c_node = to->find_context_node();

    // ASSERT((!(to_c_node->negated)));
    if (to_c_node->negated) {
        *(Cor_exEnv->err_strm) << 
                "Cycle of negative dependencies detected!\n";
    }

    if (!(Cor_subsumes_tuple(from, to, Cor_exEnv->dbg_ordsearch))) {
        from_c_node->merge_nodes(to_c_node);
        unlink_node(to_c_node);
        delete to_c_node;
    } else {
        // no merging takes place

        delete_id_tuple(to->owner_rel,to);

        unlink_node(to_c_node);
        to_c_node->free_all();
        // TMP!!
        delete to_c_node;
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_ordsearch) {
            print_all(Cor_exEnv->trace_strm);
        }
#    endif
}


// *************************************************************************
void Cor_Context::print_all(ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT((last->next == NULL));
    *file << "OrdSearch: printing state of Context.\n";
    *file << "\t---------------------------\n";
    for (Cor_ContextNode *temp = first; temp != NULL; temp = temp->next) {
        temp->print_all(file);
    }
    *file << "\t---------------------------\n";
}

// *************************************************************************
void Cor_Context::free_all()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    first = NULL;
    last = NULL;
    first_marked = NULL;
    last_marked = NULL;

    Cor_ContextNode *cur, *prev = NULL;

    for (cur = all_nodes; cur; cur=cur->all_next) {
        delete prev;
        prev = cur;
    }
    delete prev;

    all_nodes = NULL;
}

// *************************************************************************
